
#define LOAD_WITH_GLAD 1
#define GLFW_INCLUDE_NONE
#include "etc/glad.c"
#include <glad/glad.h>
#include <GLFW/glfw3.h>

#include "console.h"
#include <chrono>
#include <cmath>
//https://learnopengl.com/Getting-started/Hello-Triangle

#include "spinsleep.cpp"

using frame_clock = std::chrono::steady_clock;
GLFWwindow* window;
uint64_t vblank_phase = 0; //can't use frame_clock::time_point because compiler complains that it's not trivially copyable.
uint64_t vblank_period = 16666485; //same, no frame_clock::duration allowed
uint32_t window_width = 800;
uint32_t window_height = 600;

uint64_t div_ceil(uint64_t a, uint64_t b) { return (a + b - 1) / b; };

int64_t div_floor(int64_t a, uint64_t b) {
	if (a < 0) return -1 - (-1 - a) / b;
	return a / b;
};

void key_callback(GLFWwindow* window, int key, int scancode, int action, int mods) {
	if (key == GLFW_KEY_ESCAPE && action == GLFW_PRESS)
		glfwSetWindowShouldClose(window, true);
}

bool right_click_dragging = false; //in an attempt to find out why frames at the top are being skipped. it's not the Fence. it's not the timepoint either.
uint32_t last_drag_x_position;
uint32_t last_drag_y_position;

void mouse_button_callback(GLFWwindow* window, int button, int action, int mods) {
	double xpos, ypos;
	glfwGetCursorPos(window, &xpos, &ypos);
	if (button == GLFW_MOUSE_BUTTON_RIGHT && action == GLFW_PRESS) {
		last_drag_x_position = xpos;
		last_drag_y_position = ypos;
		right_click_dragging = true;
	}
	if (button == GLFW_MOUSE_BUTTON_RIGHT && action == GLFW_RELEASE) {
		right_click_dragging = false;
		vblank_period *= std::exp(((int64_t)last_drag_x_position - (int)xpos) / (double)window_width * 0.1);
		vblank_phase -= (int64_t)(vblank_period * (int)((int64_t)last_drag_y_position - (int)ypos)) / (int64_t)window_height;
	}
}
void mouse_cursor_callback(GLFWwindow* window, double xpos, double ypos) {
	if (right_click_dragging) {
		vblank_period *= std::exp(((int64_t)last_drag_x_position - (int)xpos) / (double)window_width * 0.1);
		vblank_phase -= (int64_t)(vblank_period * (int64_t)((int64_t)last_drag_y_position - (int)ypos)) / (int64_t)window_height;
		last_drag_x_position = (int)xpos;
		last_drag_y_position = (int)ypos;
	}
}

//we can't combine with compile_shader because of glGet*iv and glGet*InfoLog
void check_program_status(GLuint id, GLenum e) {
	GLint error_status;
	glGetProgramiv(id, e, &error_status);
	if (error_status == GL_FALSE) {
		char error_log[512];
		glGetProgramInfoLog(id, 512, 0, error_log);
		error("error while linking OpenGL program", error_log);
	}
}
void compile_shader(GLuint shader) {
	glCompileShader(shader);
	GLint error_state_of_shader;
	glGetShaderiv(shader, GL_COMPILE_STATUS, &error_state_of_shader);
	if (error_state_of_shader == GL_FALSE) {
		outc("error while compiling shader"); //comes before the shader log in case the shader log command fails

		//GLchar error_log[max_length]; //VLA causes stack overflow if max_length is too large, and templates (i.e. error()) don't work with variable-length arrays.
		char error_log[512]; //max_length 512 instead of getting the precise number. this lets us avoid vectors.
		glGetShaderInfoLog(shader, 512, 0, error_log);
		error(error_log);
	}
}

void render_loop() {
	glfwMakeContextCurrent(window);
	glfwSwapInterval(0); //doesn't really seem to do anything with glFlush
	if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress)) {
		error("Failed to initialize GLAD");
	}

	const char* vertex_shader_source = R"(#version 330 core
layout (location = 0) in vec2 pos;
layout (location = 1) in vec3 vertex_color;
out vec3 pixel_color;

void main()
{
	gl_Position = vec4(pos.x, pos.y, 0.0, 1.0);
	pixel_color = vertex_color;
})";
	const char* fragment_shader_source = R"(#version 330 core
out vec4 FragColor;
in vec3 pixel_color;

void main()
{
FragColor = vec4(pixel_color, 1.0);
})";

	GLuint vertex_shader = glCreateShader(GL_VERTEX_SHADER);
	GLuint fragment_shader = glCreateShader(GL_FRAGMENT_SHADER);
	glShaderSource(vertex_shader, 1, &vertex_shader_source, nullptr);
	glShaderSource(fragment_shader, 1, &fragment_shader_source, nullptr);
	compile_shader(vertex_shader);
	compile_shader(fragment_shader);
	GLuint shader_program = glCreateProgram();
	glAttachShader(shader_program, vertex_shader);
	glAttachShader(shader_program, fragment_shader);
	glLinkProgram(shader_program);
	check_program_status(shader_program, GL_LINK_STATUS);
	glDetachShader(shader_program, vertex_shader);
	glDetachShader(shader_program, fragment_shader);
	glDeleteShader(vertex_shader);
	glDeleteShader(fragment_shader);

	const int triangle_count = 2;
	float vertices[triangle_count * 15] = {}; //format: {x, y, luminance3}, repeat.

	unsigned int VBO, VAO;
	glGenVertexArrays(1, &VAO);
	glBindVertexArray(VAO);

	glGenBuffers(1, &VBO);
	glBindBuffer(GL_ARRAY_BUFFER, VBO);
	glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_DYNAMIC_DRAW);

	glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 20, (void*)0);
	glEnableVertexAttribArray(0);

	glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 20, (void*)(2 * sizeof(float)));
	glEnableVertexAttribArray(1);

	glClearColor(0.1f, 0.1f, 0.6f, 1.0f);
	glUseProgram(shader_program);

	float movex = 0;
	uint64_t next_frame_vblank = 0;
	auto bar = std::array<float, 3>{0.2, 0.4, 1};
	while (!glfwWindowShouldClose(window)) {

		glClear(GL_COLOR_BUFFER_BIT); //lowers fps from 760 to 580. but, if I don't do this, alt-tabbing back in doesn't work. why?
		if ((movex += 0.02f) > 0.99f) {
			movex = -1;
		}
		float nextvertices[] = {movex, -1.f, bar[0], bar[1], bar[2], movex, 1.f, bar[0], bar[1], bar[2], movex + 0.1f, -1.f, bar[0], bar[1], bar[2], movex + 0.1f, -1.f, bar[0], bar[1], bar[2], movex, 1.f, bar[0], bar[1], bar[2], movex + 0.1f, 1.f, bar[0], bar[1], bar[2]};

		glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(nextvertices), nextvertices);
		glDrawArrays(GL_TRIANGLES, 0, triangle_count * 3);

		glFinish();
		accurate_sleep_until(next_frame_vblank);
		glfwSwapBuffers(window);
		bar = std::array<float, 3>{0.2, 0.4, 1};
		uint64_t time_after_render = frame_clock::now().time_since_epoch().count();
		if (time_after_render - next_frame_vblank > 1000000) {
			outc("missed frame jump", (time_after_render - next_frame_vblank)/ 1000000);
			bar = std::array<float, 3>{0.9, 0.4, 0.4};
		}

		//move phase to just before the current timepoint, so that changing period doesn't move the tearline drastically
		vblank_phase += div_floor(time_after_render - vblank_phase, vblank_period) * vblank_period;
		next_frame_vblank = vblank_phase + div_ceil(time_after_render - vblank_phase - vblank_period / 2, vblank_period) * vblank_period + vblank_period;
		glfwPollEvents();
	}
}

int main() {
	glfwInit();
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
	glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);

	GLFWmonitor* primary = glfwGetPrimaryMonitor();
	//future: monitor callbacks for changing monitors. also, multiple monitors. and resizing windows
	const GLFWvidmode* mode = glfwGetVideoMode(primary);
	window_width = mode->width;
	window_height = mode->height;
	window = glfwCreateWindow(window_width, window_height, "Vsync test", primary, nullptr);
	if (window == NULL) {
		outc("Failed to create GLFW window");
		glfwTerminate();
		return -1;
	}
	glfwSetCursorPosCallback(window, mouse_cursor_callback);
	glfwSetKeyCallback(window, key_callback);
	glfwSetMouseButtonCallback(window, mouse_button_callback);
	render_loop();

	glfwTerminate();
}
